//////////////////////////////////////////////////////////////////////
// Simple ARM7 stub (sends RTC, TSC, and X/Y data to the ARM 9)
// -- joat
// -- modified by Darkain and others
// -- added serial test code (natrium42)
//////////////////////////////////////////////////////////////////////

 
#include <NDS/NDS.h>
 
#include <NDS/ARM7/BIOS.h>
#include <NDS/ARM7/touch.h>
#include <NDS/ARM7/clock.h>

#include <NDS/card.h>

//////////////////////////////////////////////////////////////////////

static char ser[32];
void serialReceiveHandler(char c);
void serialRead(char * a, unsigned int start);

#define EepromWaitBusy()	while (CARD_CR1 & /*BUSY*/0x80);

//#define BV(b,pos,length) ((b & (1 << pos)) != 0 ? (0xFF >> (8-length)) : 0x00);

inline char BV(char b, const int pos, const int length) {
	return ((b & (1 << pos)) != 0 ? (0xFF >> (8-length)) : 0x00);
}

inline char ISSET(char b, const int pos) {
	return ((b & (1 << pos)) != 0 ? 1 : 0);
}

void serialWriteHalfDuplex(char b) {
	IME = 0; // start of critical section
	//////////////////////////////////////////////////
	// value        0        1        2        3    //
	// freq.     4.19MHz, 2.10MHz, 1.05MHz,  524kHz //
	//////////////////////////////////////////////////
	// at freq. = 524kHz each serial bit corresponts to 9 bits for 57600bps
	CARD_CR1 = /*E*/0x8000 | /*SEL*/0x2000 | /*MODE*/0x40 | 3;	// freq = 524 kHz
	CARD_EEPDATA = 0x00; EepromWaitBusy();								// 8 parts of start bit
	CARD_EEPDATA = 0x80 | BV(b, 0, 7); EepromWaitBusy();				// 1 part  of start bit, 7 parts of b0
	CARD_EEPDATA = BV(b, 0, 2) << 6 | BV(b, 1, 6); EepromWaitBusy();	// 2 parts of b0, 6 parts of b1
	CARD_EEPDATA = BV(b, 1, 3) << 5 | BV(b, 2, 5); EepromWaitBusy();	// 3 parts of b1, 5 parts of b2
	CARD_EEPDATA = BV(b, 2, 4) << 4 | BV(b, 3, 4); EepromWaitBusy();	// 4 parts of b2, 4 parts of b3
	CARD_EEPDATA = BV(b, 3, 5) << 3 | BV(b, 4, 3); EepromWaitBusy();	// 5 parts of b3, 3 parts of b4
	CARD_EEPDATA = BV(b, 4, 6) << 2 | BV(b, 5, 2); EepromWaitBusy();	// 6 parts of b4, 2 parts of b5
	CARD_EEPDATA = BV(b, 5, 7) << 1 | BV(b, 6, 1); EepromWaitBusy();	// 7 parts of b5, 1 part  of b6
	CARD_EEPDATA = BV(b, 6, 8); EepromWaitBusy();						// 8 parts of b6
	CARD_EEPDATA = BV(b, 7, 8); EepromWaitBusy();						// 8 parts of b7
	CARD_EEPDATA = BV(b, 7, 1) << 1 | 0x7F; EepromWaitBusy();			// 1 part  of b7, 7 parts of stop bit
	CARD_EEPDATA = 0xFF; EepromWaitBusy();								// rest of stop bit
	CARD_CR1 = /*MODE*/0x40;
	IME = 1; // end of critical section
}


void serialWrite(char b) {
	IME = 0; // start of critical section
	//////////////////////////////////////////////////
	// value        0        1        2        3    //
	// freq.     4.19MHz, 2.10MHz, 1.05MHz,  524kHz //
	//////////////////////////////////////////////////
	// at freq. = 524kHz each serial bit corresponts to 9 bits for 57600bps
	CARD_CR1 = /*E*/0x8000 | /*SEL*/0x2000 | /*MODE*/0x40 | 3;	// freq = 524 kHz
	CARD_EEPDATA = 0x00; EepromWaitBusy();								// 8 parts of start bit
	ser[0] = CARD_EEPDATA;
	CARD_EEPDATA = 0x80 | BV(b, 0, 7); EepromWaitBusy();				// 1 part  of start bit, 7 parts of b0
	ser[1] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 0, 2) << 6 | BV(b, 1, 6); EepromWaitBusy();	// 2 parts of b0, 6 parts of b1
	ser[2] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 1, 3) << 5 | BV(b, 2, 5); EepromWaitBusy();	// 3 parts of b1, 5 parts of b2
	ser[3] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 2, 4) << 4 | BV(b, 3, 4); EepromWaitBusy();	// 4 parts of b2, 4 parts of b3
	ser[4] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 3, 5) << 3 | BV(b, 4, 3); EepromWaitBusy();	// 5 parts of b3, 3 parts of b4
	ser[5] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 4, 6) << 2 | BV(b, 5, 2); EepromWaitBusy();	// 6 parts of b4, 2 parts of b5
	ser[6] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 5, 7) << 1 | BV(b, 6, 1); EepromWaitBusy();	// 7 parts of b5, 1 part  of b6
	ser[7] = CARD_EEPDATA;	
	CARD_EEPDATA = BV(b, 6, 8); EepromWaitBusy();						// 8 parts of b6
	ser[8] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 7, 8); EepromWaitBusy();						// 8 parts of b7
	ser[9] = CARD_EEPDATA;
	CARD_EEPDATA = BV(b, 7, 1) << 1 | 0x7F; EepromWaitBusy();			// 1 part  of b7, 7 parts of stop bit
	ser[10] = CARD_EEPDATA;
	CARD_EEPDATA = 0xFF; EepromWaitBusy();								// rest of stop bit
	ser[11] = CARD_EEPDATA;
	CARD_CR1 = /*MODE*/0x40;

	for(unsigned int i = 0; i < 12; i++) {
		if(ser[i] != 0xFF) {
			unsigned int shift;
			for(shift = 0; shift < 8; shift++) {
				if(ser[i] & (1 << (7-shift)) == 0)
					break;
			}
			if(shift < 4)
				serialRead(&ser[i], i+1);
			else
				serialRead(&ser[i+1], i+2);
		}
	}
	IME = 1; // end of critical section
}

void serialRead(char * a, unsigned int start) {
	char b = 0x00;

	IME = 0; // start of critical section
	CARD_EEPDATA = 0x00;
	CARD_CR1 = /*E*/0x8000 | /*SEL*/0x2000 | /*MODE*/0x40 | 3;	// freq = 524 kHz
	
	for(unsigned int i = start; i < 14; i++) {
		CARD_EEPDATA = 0x00; EepromWaitBusy();
		a[i] = CARD_EEPDATA;
	}

	b = ISSET(a[1], 4) << 0 |
		ISSET(a[2], 3) << 1 |
		ISSET(a[3], 3) << 2 |
		ISSET(a[4], 2) << 3 | // 7
		ISSET(a[5], 1) << 4 | // 6
		ISSET(a[6], 1) << 5 |
		ISSET(a[7], 0) << 6 |
		ISSET(a[8], 0) << 7;


	CARD_CR1 = /*MODE*/0x40;
	IME = 1; // end of critical section

	serialWrite('>');
	serialReceiveHandler(b);
}

char serialRead(void) {
	char b = 0x00;
	char a[14];

	IME = 0; // start of critical section
	CARD_EEPDATA = 0x00;
	CARD_CR1 = /*E*/0x8000 | /*SEL*/0x2000 | /*MODE*/0x40 | 3;	// freq = 524 kHz
	
	for(unsigned int i = 0; i < sizeof(a); i++) {
		CARD_EEPDATA = 0x00; EepromWaitBusy();
		a[i] = CARD_EEPDATA;
	}

	/*
	b = ISSET(a[1], 3) << 0 |
		ISSET(a[2], 2) << 1 |
		ISSET(a[3], 1) << 2 |
		ISSET(a[4], 5) << 3 | // 7
		ISSET(a[5], 3) << 4 | // 6
		ISSET(a[6], 2) << 5 |
		ISSET(a[7], 1) << 6 |
		ISSET(a[8], 0) << 7;
	*/

	b = ISSET(a[1], 4) << 0 |
		ISSET(a[2], 3) << 1 |
		ISSET(a[3], 3) << 2 |
		ISSET(a[4], 2) << 3 | // 7
		ISSET(a[5], 1) << 4 | // 6
		ISSET(a[6], 1) << 5 |
		ISSET(a[7], 0) << 6 |
		ISSET(a[8], 0) << 7;


	CARD_CR1 = /*MODE*/0x40;
	IME = 1; // end of critical section

	return b;
}

// temporary test function:
void serialReadArray(char * a) {
	IME = 0; // start of critical section
	CARD_CR1 = /*E*/0x8000 | /*SEL*/0x2000 | /*MODE*/0x40 | 3;	// freq = 524 kHz
	
	for(unsigned int i = 0; i < 14; i++) {
		CARD_EEPDATA = 0x00; EepromWaitBusy();
		a[i] = CARD_EEPDATA;
	}

	CARD_CR1 = /*MODE*/0x40;
	IME = 1; // end of critical section
}

void serialWriteText(char * text) {
	while(*text != 0) {
		serialWrite(*text);
		text++;
	}
}

void serialWriteHex(char number) {
	for(unsigned int i = 0; i < sizeof(number) * 2; i++) {
		char nibble = (number & 0xF0) >> 4;
		if (nibble < 10)
			serialWrite('0'+nibble);
		else
			serialWrite('A'+nibble);
		number <<= 4;
	}
}

void serialWriteBin(char number) {
	for(int i = 7; i >= 0; i--) {
		if (number & (1 << i))
			serialWrite('1');
		else
			serialWrite('0');
	}
}

//////////////////////////////////////////////////////////////////////

void serialReceiveHandler(char c) {
	serialWrite(c);
}

//////////////////////////////////////////////////////////////////////








#define TOUCH_CAL_X1 (*(vs16*)0x027FFCD8)
#define TOUCH_CAL_Y1 (*(vs16*)0x027FFCDA)
#define TOUCH_CAL_X2 (*(vs16*)0x027FFCDE)
#define TOUCH_CAL_Y2 (*(vs16*)0x027FFCE0)
#define SCREEN_WIDTH    256
#define SCREEN_HEIGHT   192
s32 TOUCH_WIDTH  = TOUCH_CAL_X2 - TOUCH_CAL_X1;
s32 TOUCH_HEIGHT = TOUCH_CAL_Y2 - TOUCH_CAL_Y1;
s32 TOUCH_OFFSET_X = ( ((SCREEN_WIDTH -60) * TOUCH_CAL_X1) / TOUCH_WIDTH  ) - 28;
s32 TOUCH_OFFSET_Y = ( ((SCREEN_HEIGHT-60) * TOUCH_CAL_Y1) / TOUCH_HEIGHT ) - 28;


//////////////////////////////////////////////////////////////////////


void startSound(int sampleRate, const void* data, uint32 bytes, u8 channel=0, u8 vol=0x7F,  u8 pan=63, u8 format=0) {
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_SOURCE(channel) = (uint32)data;
  SCHANNEL_LENGTH(channel) = bytes;
  SCHANNEL_CR(channel)     = SOUND_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
}


s8 getFreeSoundChannel() {
  for (int i=0; i<16; i++) {
    if ( (SCHANNEL_CR(i) & SOUND_ENABLE) == 0 ) return i;
  }
  return -1;
}


//////////////////////////////////////////////////////////////////////


void InterruptHandler(void) {
  static int heartbeat = 0;

  if (IF & IRQ_CARD_LINE) {
	serialReceiveHandler(serialRead());


	// Test code:
	/*
    char a[14];
	serialReadArray(a);
	for(int i = 0; i < 14; i++) {
		serialWriteBin(a[i]);
		serialWrite(' ');
	}
	serialWriteText("\r\n");
	*/
  }
 
  if (IF & IRQ_VBLANK) {
    uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, batt=0, aux=0;
    int t1=0, t2=0;
    uint32 temp=0;
    uint8 ct[sizeof(IPC->curtime)];

    
    // Update the heartbeat
    heartbeat++;
 
    // Read the X/Y buttons and the /PENIRQ line
    but = XKEYS;
    if (!(but & 0x40)) {
      // Read the touch screen
      x = touchRead(TSC_MEASURE_X);
      y = touchRead(TSC_MEASURE_Y);
      xpx = ( ((SCREEN_WIDTH -60) * x) / TOUCH_WIDTH  ) - TOUCH_OFFSET_X;
      ypx = ( ((SCREEN_HEIGHT-60) * y) / TOUCH_HEIGHT ) - TOUCH_OFFSET_Y;
      z1 = touchRead(TSC_MEASURE_Z1);
      z2 = touchRead(TSC_MEASURE_Z2);
    }

    batt = touchRead(TSC_MEASURE_BATTERY);
    aux  = touchRead(TSC_MEASURE_AUX);

    // Read the time
    rtcGetTime((uint8 *)ct);
    BCDToInteger((uint8 *)&(ct[1]), 7);
 
    // Read the temperature
    temp = touchReadTemperature(&t1, &t2);
 
    // Update the IPC struct
    IPC->heartbeat = heartbeat;
    IPC->buttons   = but;
    IPC->touchX    = x;
    IPC->touchY    = y;
    IPC->touchXpx  = xpx;
    IPC->touchYpx  = ypx;
    IPC->touchZ1   = z1;
    IPC->touchZ2   = z2;
    IPC->battery   = batt;
    IPC->aux       = aux;

    for(u32 i=0; i<sizeof(ct); i++) {
      IPC->curtime[i] = ct[i];
    }

    IPC->temperature = temp;
    IPC->tdiode1 = t1;
    IPC->tdiode2 = t2;


    //sound code  :)
    TransferSound *snd = IPC->soundData;
    IPC->soundData = 0;
    if (snd) {
      for (int i=0; i<snd->count; i++) {
        s8 chan = getFreeSoundChannel();
        if (chan >= 0) {
          startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan, snd->data[i].vol, snd->data[i].pan, snd->data[i].format);
        }
      }
    }
  }
 
  // Acknowledge interrupts
  IF = IF;
}
 

//////////////////////////////////////////////////////////////////////
 

int main(int argc, char ** argv) {
  // Reset the clock if needed
  rtcReset();

  //enable sound
  SOUND_CR = SCHANNEL_ENABLE | SOUND_VOL(0x7F);
  IPC->soundData = 0;

  serialWriteText("Hello World!\r\n");
 
  // Set up the interrupt handler
  IME = 0;
  IRQ_HANDLER = &InterruptHandler;
  IE = IRQ_VBLANK | IRQ_CARD_LINE;
  IF = ~0;
  DISP_SR = DISP_VBLANK_IRQ;
  IME = 1;

  // Keep the ARM7 out of main RAM
  while (1) {
	  swiWaitForVBlank();
  }
  return 0;
}

 
//////////////////////////////////////////////////////////////////////

